/*
 * File         : BV_AudioApp.C
 * Author       : Tomasz Stachowiak
 * Created      : Fri Jul 24 18:15:10 1998
 * Revision     : $Revision$
 * Updated      : $Date$
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1998
 */

#include "adpcm.h"
#include "gsm.h"

#include "BV_Application.h"
#include "BV_AudioPlay.h"
#include "BV_AudioApp.h"
#include "BV_AudioCapThread.h"

BV_AudioApp::BV_AudioApp() : BV_Application(12567)
{
  int i;

  BV_Session_Control *ctrl = BV_Session_Control::instance();
  main_form_ = ctrl->getMainForm();
  ctrl->destroy();

  BV_AudioPlay::init();
  BV_AudioCapture::init();

  for(i = 0; i < 100; i++)
    ports[i] = NULL;

  gsm_handle_ = gsm_create();

  state_ = START;
  compression_ = PCM;

  mic_port_ = new BV_AudioCapture();
  mic_port_->open();

  cap_thread_ = new BV_AudioCapThread(this);

  cap_thread_->start();
}

BV_AudioApp::~BV_AudioApp()
{
  int i;

  cap_thread_->stop();
  sleep(1);
  delete cap_thread_;

  mic_port_->close();
  delete mic_port_;

  for(i = 0; i < 100; i++)
    if(ports[i] != NULL)
      {
	ports[i]->close();
	delete ports[i];
	ports[i] = NULL;
      }

  gsm_destroy(gsm_handle_);
}

int BV_AudioApp::process_message(RTPheader *header, char *body, int len)
{
  int sender = header->ssrc;

  if(header->type != 2)
    {
      return -1;
    }

  if(ports[sender] == NULL)
    {
      ports[sender] = new BV_AudioPlay();
      ports[sender]->open();
    }

  switch (header->pt)
    {
    case ADPCM:
      playADPCM(sender, body, len);
      break;
      
    case GSM:
      playGSM(sender, body, len);
      break;
       
    case PCM:
      ((MainForm *)main_form_)->setSpeakerLevel(get_level(body, len/2));
      ports[sender]->play(body, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);
      break;
      
    case STOP:
      ((MainForm *)main_form_)->setSpeakerLevel(0);
      if(ports[sender] != NULL)
	{
	  ports[sender]->close();
	  delete ports[sender];
	  ports[sender] = NULL;
	}
      break;
    }
  
  return 0;
}

int BV_AudioApp::playADPCM(int mbr, char *data, int len)
{
  char buffer[2*AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE];
  struct adpcm_state adpcmstate;

  memcpy(&adpcmstate.valprev, data, 2);
  adpcmstate.index = data[2];
  data += 3;
  adpcm_decoder(data, (short *)buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE, 
                &adpcmstate);

  ((MainForm *)main_form_)->setSpeakerLevel
    (get_level(buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE));

  ports[mbr]->play(buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);

  return 0;
}

int BV_AudioApp::playGSM(int mbr, char *data, int len)
{
  char buffer[AUDIO_CHUNKS_NO][320];
  int i;
  
  for(i=0; i<AUDIO_CHUNKS_NO; i++)
    {
      gsm_decode(gsm_handle_, (gsm_byte *)((gsm_frame *)data[i]), 
		 (gsm_signal *)buffer[i]);
    }
  
  ((MainForm *)main_form_)->setSpeakerLevel
    (get_level((char *)buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE));

  ports[mbr]->play((char *)buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);

  return 0;
}

int BV_AudioApp::send_audio()
{
  if(state_ == START)
    {
      switch(compression_)
	{
	case PCM:
	  sendPCM();
	  break;

	case ADPCM:
	  sendADPCM();
	  break;

	case GSM:
	  sendGSM();
	  break;
	}
    }
  else
    {
      sleep(1);
      return -1;
    }

  return 0;
}

int BV_AudioApp::sendPCM()
{
  char buffer[2*AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE];

  mic_port_->capture(buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);

  ((MainForm *)main_form_)->setMicLevel(get_level(buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE));

  send_data(PCM, buffer, 2*AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);

  return 0;
}

int BV_AudioApp::sendADPCM()
{
  char buffer[2*AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE];
  char adpcm[AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE/2 + 3];
  char *tmp;
  int size;
  
  mic_port_->capture(buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);

  ((MainForm *)main_form_)->setMicLevel(get_level(buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE));

  memcpy(adpcm, &adpcmsendstate.valprev, 2);
  adpcm[2] = adpcmsendstate.index;
  
  tmp = adpcm + 3;
  adpcm_coder((short *)buffer, tmp, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE, &adpcmsendstate);

  send_data(ADPCM, adpcm, 2*AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);

  return 0;
}

int BV_AudioApp::sendGSM()
{
  char buffer[AUDIO_CHUNKS_NO][320];
  gsm_frame frames[AUDIO_CHUNKS_NO];
  int size;
  int i;

  mic_port_->capture((char *)buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);

  ((MainForm *)main_form_)->setMicLevel(get_level((char *)buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE));

  for(i=0; i<AUDIO_CHUNKS_NO; i++)
    {
      gsm_encode(gsm_handle_, (gsm_signal *)buffer[i], (gsm_byte *)&frames[i]);
    }
  
  send_data(GSM, (char *)frames, AUDIO_CHUNKS_NO*sizeof(gsm_frame));

  return 0;
}

short BV_AudioApp::get_level(char *buffer, int no)
{
  int i;
  short max = 0;
  
  for(i = 0; i < no; i++)
    {
      if(((short *)buffer)[i] > max)
	max = ((short *)buffer)[i];
    }
  return max;
}

void BV_AudioApp::update_options(Options *options)
{
  switch(options->connection)
    {
    case _modem28:
      compression_ = GSM;
      break;

    case _isdn:
      compression_ = ADPCM;
      break;
      
    case _isdn2:
      compression_ = ADPCM;
      break;

    case _lan:
      compression_ = PCM;
      break;

    default:
      compression_ = GSM;
      break;
    }
}

void BV_AudioApp::play()
{
  state_ = START;
}

void BV_AudioApp::stop()
{
  state_ = STOP;
}
